/*
GLDM - Graphical Liberty Display Manager

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "authprovider.h"
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <pwd.h>
#include <paths.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <grp.h>
#include <err.h>
#include <stdio.h> // remove()
#include <unistd.h>
#include "utils.h"

/// \bug FIXME ASAP this must be randomized
static std::string mcookie = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa";

// docs are planted in header file
std::vector<AuthProvider*> AuthProvider::sAuthProviders;

// docs are planted in header file
AuthProvider* AuthProvider::checkAuth(const std::string& login, const std::string& password)
{
    for (std::vector<AuthProvider*>::iterator it = sAuthProviders.begin(); it != sAuthProviders.end(); it++)
    {
        AuthProvider* ap = *it;

        if (ap->checkOwnAuth(login, password))
            return ap;
    }
    return NULL;
}

// docs are planted in header file
void AuthProvider::registerProvider(AuthProvider* ap)
{
    sAuthProviders.push_back(ap);
}

// docs are planted in header file
void AuthProvider::applyEnvironment(const std::string& login, const std::string& displayname)
{
    struct passwd *pw = getpwnam(login.c_str());
    char* term = getenv("TERM");

    std::string xauthority(pw->pw_dir); xauthority += "/.Xauthority_gldmlogin";



    const int Num_Of_Variables = 10; // Number of env. variables + 1
    mChildEnv = static_cast<char**>(malloc(sizeof(char*)*Num_Of_Variables));
    int n = 0;
    if(term) mChildEnv[n++]=StrConcat("TERM=", term);
    mChildEnv[n++]=StrConcat("HOME=", pw->pw_dir);
    mChildEnv[n++]=StrConcat("SHELL=", pw->pw_shell);
    mChildEnv[n++]=StrConcat("USER=", pw->pw_name);
    mChildEnv[n++]=StrConcat("LOGNAME=", pw->pw_name);
    mChildEnv[n++]=StrConcat("PATH=", "./:/bin:/usr/bin:/usr/local/bin:/usr/X11R6/bin"); // FIXME this should be configurable: //cfg->getOption("default_path").c_str());
    mChildEnv[n++]=StrConcat("DISPLAY=", displayname.c_str());
    mChildEnv[n++]=StrConcat("MAIL=", StrConcat(_PATH_MAILDIR "/", pw->pw_name));
    mChildEnv[n++]=StrConcat("XAUTHORITY=", xauthority.c_str());
    mChildEnv[n++]=0;
    // FIXME BUG memory leak: mChildEnv not freed
}

bool AuthProvider::dropPrivileges(const std::string& login)
{
    struct passwd *pw = getpwnam(login.c_str());
    //printf("Dropping privileges to %d...\n", pw->pw_uid);
    if ((initgroups(pw->pw_name, pw->pw_gid) != 0))
        err(1, "initgroups()");
    if ((setgid(pw->pw_gid) != 0))
        err(2, "setgid()");
    if ((setuid(pw->pw_uid) != 0))
        err(3, "setuid()");
    return true;
}

void AuthProvider::setClientAuth(const char* login, const char* mcookie, const char* displayname)
{
    struct passwd *pw = getpwnam(login);
    int r;
    std::string home(pw->pw_dir);
    std::string authfile = home + "/.Xauthority_gldmlogin"; // FIXME perhaps somewhere else?
    remove(authfile.c_str());
    // FIXME BUG don't use system
    // see http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=529306
    std::string cmd = std::string("/usr/bin/xauth") + " -q -f " + authfile + " add " + displayname + " . " + mcookie;
    r = system(cmd.c_str());
}

// docs are planted in header file
bool AuthProvider::switchUser(const std::string& login, const std::string& command, const std::string& displayname, const std::string& sessioncommand)
{
    pid_t proc;
    if ((proc=fork())==0)
    {
        if (!dropPrivileges(login)) {
            warnx("dropping privileges failed");
            return false;
        }
        applyEnvironment(login, displayname);
        setClientAuth(login.c_str(), mcookie.c_str(), displayname.c_str());

        execle("/bin/sh", "sh", "-c", sessioncommand.c_str(), (char*)0, mChildEnv);
        return true; // never happens
    } else {
        int returninfo;
        wait(&returninfo);
        printf("Finished with return value %d\n",returninfo);
        return true;
    }
}

// docs are planted in header file
std::vector<std::string> AuthProvider::getLoginListUsers()
{
    std::vector<std::string> lst;
    for (std::vector<AuthProvider*>::iterator it = sAuthProviders.begin(); it != sAuthProviders.end(); it++)
    {
        std::vector<std::string> v((*it)->getOwnLoginListUsers());
        lst.insert(lst.end(), v.begin(), v.end());
        // FIXME remove duplicates
    }

    return lst;
}

std::string AuthProvider::getDisplayName(std::string username)
{
    std::vector<std::string> lst;
    for (std::vector<AuthProvider*>::iterator it = sAuthProviders.begin(); it != sAuthProviders.end(); it++)
    {
        std::string dn = (*it)->getOwnDisplayName(username);
        if (dn!="")
        {
            return dn;
        }
    }
    return username;
}
std::string AuthProvider::getHomeDir(std::string username)
{
    std::vector<std::string> lst;
    for (std::vector<AuthProvider*>::iterator it = sAuthProviders.begin(); it != sAuthProviders.end(); it++)
    {
        std::string hd = (*it)->getOwnHomeDir(username);
        if (hd!="")
        {
            return hd;
        }
    }
    return username;
}

std::string AuthProvider::getOwnDisplayName(std::string username)
{
    return "";
}
std::string AuthProvider::getOwnHomeDir(std::string username)
{
    return "";
}
